/****************************************************************************

FILE	povslc.c

PURPOSE 
    This file contains main function and setup code for shading language 
    compiler for POV-Ray (povslc)

*****************************************************************************/
/*****************************************************************************                                                                              
 Copyright (c) 1995
      Robert J. Mercier
 Copyright 2000
      Vahur Krouverk

THIS CODE IS SUPPLIED ``AS IS'' WITH NO WARRANTY EXPRESSED
OR IMPLIED.  NO CLAIMS IS MADE TO IT'S SUITABILITY FOR
ANY PURPOSE AND NO RESPONSIBILITY IS ACCEPTED FOR ANY DAMAGES
OF ANY KIND RESULTING FROM IT'S USE.

When (if) this code is included into POV the usual POV legal stuff will apply.

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <assert.h>

#include "shade.h"
#include "slgram.h"
#include "shmem.h"
#include "shdrglob.h"

int cflag;
int check_flag;

extern FILE * yyin;

void parse_args(int argc, char * argv[]);
void output_opts(void);
int preprocess_input(void);
void close_temp_file(void);

static char * input_filename = NULL;
static char * output_filename = NULL;
static char * tmp_filename = NULL;
static char * user_prep_args = NULL;
char * cur_input_file=NULL;
FILE * input_file = NULL;
FILE * output_file = NULL;
FILE * tmp_file = NULL;

int main(int argc, char *argv[])
{
    
    int error_code=0;
    
    /* register memory releasing function */
    atexit(sh_release_mem);
    atexit(close_temp_file);
    
    fprintf(stderr, "%s %d.%d\n", SHADER_PRODUCT_STRING, SHADER_VERSION_MAJOR,
                SHADER_VERSION_MINOR);    
    parse_args(argc, argv);
    
    if (input_filename == NULL){
        fprintf(stderr, "Error: No input file given!\n");
        exit(1);
    }
    cur_input_file = strsave(input_filename);
    error_code = preprocess_input();
    if (error_code != 0)
        exit(1);

    if (output_filename == NULL){
        /*
            output file not given by user. Deduce output file name from
            input file's base name and add extension .slp
        */
        char *point_pos = strrchr(input_filename, '.');
        char buf[FILENAME_MAX];
        if (point_pos != NULL){
            int offset = point_pos-input_filename;
            strcpy(buf, input_filename);
            strcpy(buf+offset+1,"slp");
        }
        else{
            strcpy(buf, input_filename);
            strcat(buf, ".slp");
        }
        output_filename = strsave(buf);
    }
    if (strcmp(output_filename, input_filename) == 0){
        /* sanity check to avoid losing of input file */
        fprintf(stderr, "Error: Input and output files have same names.\n");
        exit(1);
    }
    output_file = fopen(output_filename, "w");
    if (output_file == NULL){
        fprintf(stderr, "Error: Can't open output file %s\n", output_filename);
        exit(1);
    }
    
    fprintf(stderr, "Compiling %s...\n", input_filename);
    
    sym_initialize();
    yyparse();
    s_fixcode(&code[0], get_dot());
    s_fixvars(&code[0], get_dot());

    dumpsyms();
    if (cflag)
        s_dumpobj(&code[0], get_dot());
    s_dumpcode(&code[0], get_dot());
    
    fprintf(stderr, "Done. Result is in file %s\n\n", output_filename);
    if (input_file != NULL){
        fclose(input_file);
    }
    fclose(output_file);
    
    /* 
      temporary file is closed in function close_temporary_file at program exit.
      No need to handle it here.
    */
    return 0;
}


void parse_args(int argc, char * argv[])
{
    int c;
    int count;
    
    if (argc == 1){
        output_opts();
    }
    for (count = 1; count < argc; count ++){
        int offs = 0;
        if (argv[count][0] == '-'){
            switch(argv[count][1]){
                case '?':
                case 'h':
                    output_opts();
                    exit(1);
                case 'c':
                    cflag++;                                                        
                    break;                                                          
                case 'f':                                                           
                    output_functions = 1;                                           
                    break;                                                          
                case 'o':
                    output_filename = strsave( &(argv[count][2]));
                    break;
                case 'w':
                    warning_flag = 1;
                    break;
                case 'v':
                    check_flag = 1;
                    break;
                case 'd':
                    {
                        extern int yydebug;
                        yydebug = 1;
                    }
                    break;
                case '-':
                case '/':
                    offs++;
                    /* 
                        note: deliberate fallthrough to default: pass 
                        --option to preprocessor as -option
                     */
                default:
                    if (user_prep_args == NULL){
                        user_prep_args = strsave( &(argv[count][offs]));
                    }
                    else{
                        char *preps = user_prep_args;
                        user_prep_args = sh_malloc(strlen(user_prep_args) + 
                                    strlen(&argv[count][offs]) + 2);
                        strcpy(user_prep_args, preps);
                        strcat(user_prep_args," ");
                        strcat(user_prep_args, &argv[count][offs]);
                        sh_free(preps);
                    }
                    break;
            }
        }
        else{
            /* input file name */
            input_filename = strsave(&(argv[count][0]));
        }
    }
}

void output_opts(void)
{
    fprintf(stderr, "Shading language compiler for POV-Ray.\n");
    fprintf(stderr, "Usage:\n");
    fprintf(stderr, "%s [opts] input_file\n", SHADER_PRODUCT_STRING);
    fprintf(stderr, "options:\n");
    fprintf(stderr, "        -?, -h      : this help screen.\n");
    fprintf(stderr, "        -w          : display warnings.\n");
    fprintf(stderr, "        -v          : generate check ops in code\n");
    fprintf(stderr, "        -c          : output machine opcodes.\n");
    fprintf(stderr, "        -f          : output functions and machine opcodes.\n");
    fprintf(stderr, "                      (no code generated!)\n");
    fprintf(stderr, "        -d          : turns on yacc parser tracing to error stream\n");
    fprintf(stderr, "        -o filename : specify output file.\n");
    fprintf(stderr, "                      (default is [inputfile].slp).\n");
    fprintf(stderr, "        --XXXX      : pass option -XXXX to preprocessor\n");
    fprintf(stderr, "                      (useful, if there is clash between options)\n");
    fprintf(stderr, " all other options are passed to preprocessor\n");
    fprintf(stderr, "\n\n");
}


int preprocess_input(void)
{
    char * prep = getenv("POVSLCPP");
    char * prep_param = getenv("POVSLCPP_PARAM");
    if (prep != NULL){
        /* 
            preprocess shader file by constructing preprocessor running string
            and sending it to command()
        */
        char *startstr;
        int startstrlen;
        int rval;
        tmp_filename = tmpnam(NULL);
        if (tmp_filename == NULL){
            fprintf(stderr, "Error: Can't create temporary file name\n");
        }
        /* create shell command for preprocessing */
        startstrlen = strlen(prep) + strlen(tmp_filename) + strlen(input_filename);
        if (prep_param != NULL)
            startstrlen += strlen(prep_param);
        if (user_prep_args != NULL)
            startstrlen += strlen(user_prep_args);
        startstrlen += 5;
        startstr = sh_malloc(startstrlen);
        strcpy(startstr, prep);
        strcat(startstr, " ");
        if (prep_param != NULL){
            strcat(startstr, prep_param);
            strcat(startstr," ");
        }
        if (user_prep_args != NULL){
            strcat(startstr, user_prep_args);
            strcat(startstr, " ");
        }
        strcat(startstr, input_filename);
        strcat(startstr," ");
        strcat(startstr, tmp_filename);
        assert(strlen(startstr) < startstrlen);
        
        /* invoke preprocessor */
        fprintf(stderr, "Preprocessing %s ...\n", input_filename);
        rval = system(startstr);
        fprintf(stderr,"Done.\n");
        if (rval != 0){
            fprintf(stderr, "Error: Preprocessor returned %d\n",rval);
            fprintf(stderr, "Preprocessor was started with following command line:\n");
            fprintf(stderr, "%s\n",startstr);
            return (1);
        }
        tmp_file = fopen(tmp_filename, "r");
        yyin = tmp_file;
    }
    else{
        /* preprocessor not specified. Use input file directly */
        input_file = fopen(input_filename, "r");
        if (input_file == NULL){
            fprintf(stderr, "Error: Can't open input file %s\n", input_filename);
            return (1);
        }
        yyin = input_file;
    }
    return 0;
}

/*
    Called during program exit (registered with atexit).
    Exit function is used during syntax errors, therefore we should catch
    program termination from exit function in order to get rid of temporary 
    file.
*/
void close_temp_file(void)
{
    if (tmp_file != NULL){
        fclose(tmp_file);
        remove(tmp_filename);
        tmp_file = NULL;
    }
}

