/****************************************************************************

FILE	shdrexec.c

PURPOSE 
    	This file contains implementation for shader's stack-based virtual 
	machine
	
*****************************************************************************/
/*****************************************************************************                                                                              
 Copyright (c) 1995
      Robert J. Mercier
 Copyright 2000
      Vahur Krouverk

THIS CODE IS SUPPLIED ``AS IS'' WITH NO WARRANTY EXPRESSED
OR IMPLIED.  NO CLAIMS IS MADE TO IT'S SUITABILITY FOR
ANY PURPOSE AND NO RESPONSIBILITY IS ACCEPTED FOR ANY DAMAGES
OF ANY KIND RESULTING FROM IT'S USE.

When (if) this code is included into POV the usual POV legal stuff will apply.

*****************************************************************************/
#include "frame.h"
#ifdef ShaderPatch

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>

#include "shdrvars.h"
#include "opcodes.h"
#include "colour.h"
#include "vector.h"
#include "texture.h"
#include "render.h"
#include "ray.h"
#include "povproto.h"
#include "povray.h"
#include "lighting.h"

static DBL spline_f(DBL t, int npnts, int type, DBL * etop);
static void  spline_p(DBL *c, DBL t, int npnts, int type,  DBL *etop);
static void reflect( DBL result[3], DBL I[3], DBL N[3]);
static void refract(DBL result[3], DBL I[3], DBL N[3], DBL eta);
static void fresnel(DBL I[3], DBL N[3], DBL eta, DBL *pkr, DBL *pkt);
static DBL cellnoise(int dim, DBL vector[3], DBL time);
static void cellnoise3d(int dim, DBL result[3], DBL vector[3], DBL time);
static DBL pnoise1(int dim, DBL point[3], DBL time, 
                        DBL pperiod[3], DBL tperiod);
static void pnoise3(int dim, DBL resultp[3],
                      DBL point[3], DBL time,
                      DBL pperiod[3], DBL tperiod);
static void rotate(DBL resultp[3], DBL point[3], DBL angle, DBL startv[3],
        DBL endv[3]);
static DBL ptlen(DBL p1[3], DBL p2[3], DBL pq[3]);
static void shader_trace(shader_inst_t *pi, DBL color[3], DBL startp[3], DBL direction[3]);
static void ambient(DBL *c, shader_inst_t * p_shader);
static void phong(DBL *p, DBL *n, DBL *e, DBL size, shader_inst_t *pi, DBL *c);
static void specular(DBL *p, DBL *n, DBL *e, DBL roughness, shader_inst_t *pi, DBL *c);
static void specularbrdf(DBL *p, DBL *n, DBL *e, DBL roughness, DBL *c);
static void diffuse(DBL *p, DBL *n, shader_inst_t *pi, DBL *c);
static void s_prf(char *fmt, int nargs, DBL *etop);
static int color_space_conv(shader_inst_t *pi, int offset);
static int point_space_conv(shader_inst_t *pi, int offset);
static int spline_type_conv(shader_inst_t *pi, int type);

static int do_illuminance(VECTOR Intersection, LIGHT_SOURCE **Light,
    COLOUR Light_Colour, VECTOR Cone_Axis, DBL Cone_Angle, RAY * Eye);
static int in_illuminance();

int func_param_count;


/*
    Called after shader initialization on opcode ENDINIT to set user specified 
    parameter values.
*/
int s_prepexec(shader_inst_t *pi, pov_arg_t *args)
{
    pov_arg_t *a;
    
    for(a = args; a != NULL; a = a->next){
        switch (a->type) {
            case T_FLOAT:
                pi->data[a->slot] = a->v.f;
                break;
            case T_COLOR:
            case T_VECTOR:
                pi->data[a->slot  ] = a->v.v[X];
                pi->data[a->slot+1] = a->v.v[Y];
                pi->data[a->slot+2] = a->v.v[Z];
                break;
        }
    }
    return args != NULL;
}

#define TEST_BOUNDS 0


#define ARRAY_SUBSCRIPT3 \
    if (array_index){\
        if (array_index >= s_getdim(p, slot) || array_index < 0){\
            Error("Shader array subscription error: index [%d] @%d\n", \
                    array_index, cia-1);\
        }\
        slot += array_index*3;\
        array_index = 0;\
    }

#define ARRAY_SUBSCRIPT \
    if (array_index){\
        if (array_index >= s_getdim(p, slot) || array_index < 0){\
            Error("Shader array subscription error: index [%d]  @%d\n", \
                    array_index, cia-1);\
        }\
        slot += array_index;\
        array_index = 0;\
    }


#define GETSLOT ((codeseg[cia]<<8)|codeseg[cia+1])
#define GETSLOT2 ((codeseg[cia+2]<<8)|codeseg[cia+3])
#define MAXSTACK    128

void s_execute(shader_inst_t *pi)
{
    DBL e_stack[MAXSTACK];
    VECTOR noiz, vec_result, v;
    DBL f, x, y, z;
    shader_t *p = pi->shader;
    DBL *pf;
    DBL *etop = &e_stack[MAXSTACK];
    unsigned char * codeseg = p->code;
    DBL *dataseg = pi->data;
    int noiz_dim;
    int cia= pi->inst_startaddr;
    int cond, slot, i, j;
    int array_index=0;
    static int srand_init = 0;
    unsigned char op;
    while ((op = codeseg[cia++]) != oEND) {
        switch (op) {
        case oENDINIT:
            /* skip argument default value init in next occurrence */
            if (p->initialized == 0){
                p->startaddr = cia-1;
                memcpy(p->data_templ, dataseg, sizeof(DBL)*p->dsize);                        
                p->initialized++;
            }
            if (s_prepexec(pi, pi->args) == 0){
                /* no arguments present, skip argument init */
                pi->inst_startaddr = cia;
            }
            else{
                pi->inst_startaddr = p->startaddr;
            }
            break;
        case oARR:
            array_index = *etop++;
            break;
        case oARR3:
            array_index = *etop++;
            break;
        case oASSIGN3:
            slot = GETSLOT;
            ARRAY_SUBSCRIPT3;
            dataseg[slot  ] = *etop++;
            dataseg[slot+1] = *etop++;
            dataseg[slot+2] = *etop++;
            cia += 2;
            break;
        case oASSIGN:
            slot = GETSLOT;
            ARRAY_SUBSCRIPT;
            dataseg[slot  ] = *etop++;
            cia += 2;
            break;
        case oASSIGNS:
            slot = GETSLOT;
            i = *etop++;
            pi->inst_strings[slot].string = pi->inst_strings[i].string;
            cia += 2;
            break;
        case oADDEQ3:
            slot = GETSLOT;
            ARRAY_SUBSCRIPT3;
            dataseg[slot  ] += *etop++;
            dataseg[slot+1] += *etop++;
            dataseg[slot+2] += *etop++;
            cia += 2;
            break;
        case oADDEQ:
            slot = GETSLOT;
            ARRAY_SUBSCRIPT;
            dataseg[slot  ] += *etop++;
            cia += 2;
            break;
        case oSUBEQ3:
            slot = GETSLOT;
            ARRAY_SUBSCRIPT3;
            dataseg[slot  ] -= *etop++;
            dataseg[slot+1] -= *etop++;
            dataseg[slot+2] -= *etop++;
            cia += 2;
            break;
        case oSUBEQ:
            slot = GETSLOT;
            ARRAY_SUBSCRIPT;
            dataseg[slot  ] -= *etop++;
            cia += 2;
            break;
        case oMULEQ3:
            slot = GETSLOT;
            ARRAY_SUBSCRIPT3;
            dataseg[slot  ] *= *etop++;
            dataseg[slot+1] *= *etop++;
            dataseg[slot+2] *= *etop++;
            cia += 2;
            break;
        case oMULEQ:
            slot = GETSLOT;
            ARRAY_SUBSCRIPT;
            dataseg[slot  ] *= *etop++;
            cia += 2;
            break;
        case oDIVEQ3:
            slot = GETSLOT;
            ARRAY_SUBSCRIPT3;
            dataseg[slot  ] /= *etop++;
            dataseg[slot+1] /= *etop++;
            dataseg[slot+2] /= *etop++;
            cia += 2;
            break;
        case oDIVEQ:
            slot = GETSLOT;
            ARRAY_SUBSCRIPT;
            dataseg[slot  ] /= *etop++;
            cia += 2;
            break;
        case oCF2T:
            *(etop-1) = *etop;
            --etop;
            *(etop-1) = *etop;
            --etop;
            break;
        case oADD3:
            *(etop+3) = *(etop+3) + *(etop  );
            *(etop+4) = *(etop+4) + *(etop+1);
            *(etop+5) = *(etop+5) + *(etop+2);
            etop += 3;
            break;
        case oADD:
            *(etop+1) = *(etop+1) + *(etop  );
            ++etop;
            break;
        case oSUB3:
            *(etop+3) = *(etop+3) - *(etop  );
            *(etop+4) = *(etop+4) - *(etop+1);
            *(etop+5) = *(etop+5) - *(etop+2);
            etop += 3;
            break;
        case oSUB:
            *(etop+1) = *(etop+1) - *(etop  );
            ++etop;
            break;
        case oMUL3:
            *(etop+3) = *(etop+3) * *(etop  );
            *(etop+4) = *(etop+4) * *(etop+1);
            *(etop+5) = *(etop+5) * *(etop+2);
            etop += 3;
            break;
        case oMUL:
            *(etop+1) = *(etop+1) * *(etop  );
            ++etop;
            break;
        case oDIV3:
            *(etop+3) = *(etop+3) / *(etop  );
            *(etop+4) = *(etop+4) / *(etop+1);
            *(etop+5) = *(etop+5) / *(etop+2);
            etop += 3;
            break;
        case oDIV:
            *(etop+1) = *(etop+1) / *(etop  );
            ++etop;
            break;
        case oDOTP:
            f = *(etop  ) * *(etop+3) +
                *(etop+1) * *(etop+4) +
                *(etop+2) * *(etop+5);
            etop += 6;
            *(--etop) = f;
            break;
        case oCROSSP:
            x = *(etop+4) * *(etop+2) - *(etop+5) * *(etop+1);
            y = *(etop+5) * *(etop  ) - *(etop+3) * *(etop+2);
            z = *(etop+3) * *(etop+1) - *(etop+4) * *(etop  );
            etop += 6;
            *(--etop) = z;
            *(--etop) = y;
            *(--etop) = x;
            break;
        case oNEGATE3:
            *(etop  ) = - *(etop  );
            *(etop+1) = - *(etop+1);
            *(etop+2) = - *(etop+2);
            break;
        case oNEGATE:
            *(etop  ) = - *(etop  );
            break;
        case oPUSH3:
            slot = GETSLOT;
            ARRAY_SUBSCRIPT3;
            *(--etop) = dataseg[slot+2];
            *(--etop) = dataseg[slot+1];
            *(--etop) = dataseg[slot  ];
            cia += 2;
            break;
        case oPUSH:
            slot = GETSLOT;
            ARRAY_SUBSCRIPT;
            *(--etop) = dataseg[slot];
            cia += 2;
            break;
        case oPUSHARR:
            slot = GETSLOT;
            i = s_getdim(p, slot)-1;
            for (i; i>=0; --i){
                *(--etop) = dataseg[slot+i];
            }
            cia += 2;
            break;
        case oPUSHARR3:
            slot = GETSLOT;
            i = s_getdim(p, slot)*3-1;
            while(i>=0){
                *(--etop) = dataseg[slot+i--];
                *(--etop) = dataseg[slot+i--];
                *(--etop) = dataseg[slot+i--];
            }
            cia += 2;
            break;
        case oPUSHALEN:
            slot = GETSLOT;
            i = s_getdim(p, slot);
            *(--etop) = (DBL)i;
            cia += 2;
            break;
        case oPUSHC:
            i = GETSLOT;
            *(--etop) = (DBL) i;
            cia += 2;
            break;
        case oPUSHS:
            i = GETSLOT;
            *(--etop) = (DBL) i;
            cia += 2;
            break;
        case oSIN:
            *(etop) = sin(*(etop));
            break;
        case oASIN:
            *(etop) = asin(*(etop));
            break;
        case oCOS:
            *(etop) = cos(*(etop));
            break;
        case oACOS:
            *(etop) = acos(*(etop));
            break;
        case oTAN:
            *(etop) = tan(*(etop));
            break;
        case oATAN:
            *(etop) = atan(*(etop));
            break;
        case oATAN2:
            *(etop+1) = atan2(*(etop+1), *(etop));
            ++etop;
            break;
        case oRADIANS:
            *(etop) = M_PI * *(etop) / 180.;
            break;
        case oDEGREES:
            *(etop) = 180. * *(etop) / M_PI;
            break;
        case oSQRT:
            *(etop) = sqrt(*(etop));
            break;
        case oSQR:
            *(etop) *= *(etop);
            break;
        case oINVSQRT:
            *(etop)= 1/sqrt(*(etop));
            break;
        case oPOW:
            *(etop+1) = pow(*(etop+1), *(etop));
            ++etop;
            break;
        case oEXP:
            *(etop) = exp(*(etop));
            break;
        case oLOG:
            *(etop) = log(*(etop));
            break;
        case oLOG2:
            *(etop+1) = log(*(etop))/log(*(etop+1));
            etop++;
            break;
        case oMOD:
            *(etop+1) = fmod(*(etop+1), *(etop));
            ++etop;
            break;
        case oABS:
            *(etop) = fabs(*(etop));
            break;
        case oSIGN:
            if (*(etop) < 0.)
                *(etop) = -1.;
            else
                *(etop) = 1.;
            break;
        case oMIN:
            if (*(etop+1) > *(etop))
                *(etop+1) = *(etop);
            ++etop;
            break;
        case oMIN3:
            if (*(etop+3) > *(etop))
                *(etop+3) = *(etop);
            if (*(etop+4) > *(etop+1))
                *(etop+4) = *(etop+1);
            if (*(etop+5) > *(etop+2))
                *(etop+5) = *(etop+2);
            etop +=3;
            break;
        case oMAX:
            if (*(etop+1) < *(etop))
                *(etop+1) = *(etop);
            ++etop;
            break;
        case oMAX3:
            if (*(etop+3) < *(etop))
                *(etop+3) = *(etop);
            if (*(etop+4) < *(etop+1))
                *(etop+4) = *(etop+1);
            if (*(etop+5) < *(etop+2))
                *(etop+5) = *(etop+2);
            etop +=3;
            break;
        case oCLAMP:
            if (*(etop+2) < *(etop+1))
                *(etop+2) = *(etop+1);
            else if (*(etop+2) > *(etop))
                *(etop+2) = *(etop);
            etop += 2;
            break;
        case oCLAMP3:
            if (*(etop+8) < *(etop+5))
                *(etop+8) = *(etop+5);
            else if (*(etop+8) > *(etop+2))
                *(etop+8) = *(etop+2);
            if (*(etop+7) < *(etop+4))
                *(etop+7) = *(etop+4);
            else if (*(etop+7) > *(etop+1))
                *(etop+7) = *(etop+1);              
            if (*(etop+6) < *(etop+3))
                *(etop+6) = *(etop+3);
            else if (*(etop+6) > *(etop))
                *(etop+6) = *(etop);
            etop += 6;
            break;
        case oCEIL:
            *(etop) = ceil(*(etop));
            break;
        case oFLOOR:
            *(etop) = floor(*(etop));
            break;
        case oROUND:
            *(etop) = (long)(*(etop));
            break;
        case oRANDOM:
            if (! srand_init){
                /* 
                   initialize rand only if needed, otherwise it causes
                   considerable slowdown for whole shader 
                */
                ++srand_init;
                srand(1);
            }
            *(--etop) = ((DBL)rand())/RAND_MAX;
            break;
        case oRANDOM3:
            if (! srand_init){
                ++srand_init;
                srand(1);
            }
            *(--etop) = ((DBL)rand())/RAND_MAX;
            *(--etop) = ((DBL)rand())/RAND_MAX;
            *(--etop) = ((DBL)rand())/RAND_MAX;
            break;                        
        case oSTEP:
            if (*(etop) < *(etop+1))
                *(etop+1) = 0.;
            else
                *(etop+1) = 1.;
            ++etop;
            break;
        case oSMOOTHSTEP:
            if (*(etop) < *(etop+2))
                *(etop+2) = 0.;
            else if (*(etop) >= *(etop+1))
                *(etop+2) = 1.;
            else {
                f = (*(etop) - *(etop+2)) / (*(etop+1) - *(etop+2));
                *(etop+2) = f*f*(3-2*f);
            }
            etop += 2;
            break;
        case oCOMP:
            {
                int index=(int)*(etop);
                *(etop+3)=*(etop+1+index);
                etop+=3;
            }
            break;
        case oXCOMP:
            {
                *(etop+2)=*(etop);
                etop+=2;
            }
            break;
        case oYCOMP:
            {
                *(etop+2)=*(etop+1);
                etop+=2;
            }
            break;
        case oZCOMP:
            {
                etop+=2;
            }
            break;
        case oSETCOMP:
            {
                int offset=(int) *etop;
                slot = GETSLOT;
                ARRAY_SUBSCRIPT3;
                dataseg[slot+offset] = *(etop+1);
                etop+=2;
                cia+=2;
            }
            break;
        case oSETXCOMP:
            {
                slot = GETSLOT;
                ARRAY_SUBSCRIPT3;
                dataseg[slot] = *(etop++);
                cia+=2;
            }
            break;
        case oSETYCOMP:
            {
                slot = GETSLOT;
                ARRAY_SUBSCRIPT3;
                dataseg[slot+1] = *(etop++);
                cia+=2;
            }
            break;
        case oSETZCOMP:
            {
                slot = GETSLOT;
                ARRAY_SUBSCRIPT3;
                dataseg[slot+2] = *(etop++);
                cia+=2;
            }
            break;
        case oLENGTH:
            *(etop+2) = sqrt( *(etop  ) * *(etop  ) +
                      *(etop+1) * *(etop+1) +
                      *(etop+2) * *(etop+2));
            etop += 2;
            break;
        case oDISTANCE:
            x = *(etop+3) - *(etop  );
            y = *(etop+4) - *(etop+1);
            z = *(etop+5) - *(etop+2);
            etop += 6;
            *(--etop) = sqrt(x*x + y*y + z*z);
            break;
        case oNORMALIZE:
            f = sqrt(*(etop  ) * *(etop  ) +
                 *(etop+1) * *(etop+1) +
                 *(etop+2) * *(etop+2));
            if (f == 0.) {
                *(etop  ) = 0.;
                *(etop+1) = 0.;
                *(etop+2) = 0.;
            } else {
                *(etop  ) /= f;
                *(etop+1) /= f;
                *(etop+2) /= f;
            }
            break;
        case oMIX:
            f = *etop++;
            *(etop+1) = (1. - f) * *(etop+1) + f * *(etop  );
            etop += 1;
            break;
        case oMIX3:
            f = *etop++;
            *(etop+3) = (1. - f) * *(etop+3) + f * *(etop  );
            *(etop+4) = (1. - f) * *(etop+4) + f * *(etop+1);
            *(etop+5) = (1. - f) * *(etop+5) + f * *(etop+2);
            etop += 3;
            break;
        case oNOISE: 
            /* TODO: 4-dim noise does not work! */
            noiz_dim = GETSLOT;
            if (noiz_dim==1)
                noiz[0]=noiz[1]=noiz[2]=*(etop);
            else if (noiz_dim==2){
                noiz[0]=noiz[1]=*(etop);
                noiz[2]=*(etop+1);
            }
            else if (noiz_dim==3){
                noiz[0]=*(etop);
                noiz[1]=*(etop+1);
                noiz[2]=*(etop+2);
            }
            else if (noiz_dim==4){
                noiz[0]=*(etop);
                noiz[1]=*(etop+1);
                noiz[2]=*(etop+2);
            }
            etop+=noiz_dim-1;
            *(etop) = Noise(noiz);
            cia+=2;
            break;
        case oNOISE3:
            /* TODO: 4-dim noise does not work! */
            noiz_dim = GETSLOT;
            if (noiz_dim==1){
                noiz[0]=noiz[1]=noiz[2]=*(etop);
            }
            else if (noiz_dim==2){
                noiz[0]=noiz[1]=*(etop);
                noiz[2]=*(etop+1);
            }
            else if (noiz_dim==3){
                noiz[0]=*(etop);
                noiz[1]=*(etop+1);
                noiz[2]=*(etop+2);
            }
            else if (noiz_dim==4){
                noiz[0]=*(etop);
                noiz[1]=*(etop+1);
                noiz[2]=*(etop+2);
            }
            etop+=noiz_dim;
            DNoise(vec_result, noiz);
            *(--etop) = vec_result[Z];
            *(--etop) = vec_result[Y];
            *(--etop) = vec_result[X];
            cia+=2;
            break;
        case oCELLNOISE:
            noiz_dim = GETSLOT;
            if (noiz_dim==1){
                noiz[0] = *(etop);
            }
            else if (noiz_dim==2){
                noiz[0] = *(etop);
                noiz[1] = *(etop+1);
            }
            else if (noiz_dim==3){
                noiz[0] = *(etop);
                noiz[1] = *(etop+1);
                noiz[2] = *(etop+2);
            }
            else if (noiz_dim==4){
                noiz[0] = *(etop);
                noiz[1] = *(etop+1);
                noiz[2] = *(etop+2);
                x = *(etop+3);
            }
            etop+=noiz_dim-1;
            *(etop) = cellnoise(noiz_dim, noiz, x);
            cia+=2;
            break;
        case oCELLNOISE3:
            noiz_dim = GETSLOT;
            if (noiz_dim==1){
                noiz[0] = *(etop);
            }
            else if (noiz_dim==2){
                noiz[0] = *(etop);
                noiz[1] = *(etop+1);
            }
            else if (noiz_dim==3){
                noiz[0] = *(etop);
                noiz[1] = *(etop+1);
                noiz[2] = *(etop+2);
            }
            else if (noiz_dim==4){
                noiz[0] = *(etop);
                noiz[1] = *(etop+1);
                noiz[2] = *(etop+2);
                x = *(etop+3);
            }
            etop+=noiz_dim;
            cellnoise3d(noiz_dim, vec_result, noiz, x);
            *(--etop) = vec_result[Z];
            *(--etop) = vec_result[Y];
            *(--etop) = vec_result[X];
            cia+=2;
            break;
        case oPNOISE: 
            noiz_dim = GETSLOT;
            if (noiz_dim==1){
                v[0] = *(etop);
                noiz[0]=*(etop+1);
            }
            else if (noiz_dim==2){
                v[0] = *(etop);
                v[1] = *(etop+1);
                noiz[0]=*(etop+2);
                noiz[1]=*(etop+3);
            }
            else if (noiz_dim==3){
                v[0] = *(etop);
                v[1] = *(etop+1);
                v[2] = *(etop+2);
                noiz[0]=*(etop+3);
                noiz[1]=*(etop+4);
                noiz[2]=*(etop+5);
            }
            else if (noiz_dim==4){
                y = *(etop);
                v[0] = *(etop+1);
                v[1] = *(etop+2);
                v[2] = *(etop+3);
                x = *(etop+4);
                noiz[0]=*(etop+5);
                noiz[1]=*(etop+6);
                noiz[2]=*(etop+7);
            }
            etop+=noiz_dim*2;
            *(--etop) = pnoise1(noiz_dim, noiz, x, v, y );
            cia+=2;
            break;
        case oPNOISE3:
            noiz_dim = GETSLOT;
            if (noiz_dim==1){
                v[0] = *(etop);
                noiz[0]=*(etop+1);
            }
            else if (noiz_dim==2){
                v[0] = *(etop);
                v[1] = *(etop+1);
                noiz[0]=*(etop+2);
                noiz[1]=*(etop+3);
            }
            else if (noiz_dim==3){
                v[0] = *(etop);
                v[1] = *(etop+1);
                v[2] = *(etop+2);
                noiz[0]=*(etop+3);
                noiz[1]=*(etop+4);
                noiz[2]=*(etop+5);
            }
            else if (noiz_dim==4){
                y = *(etop);
                v[0] = *(etop+1);
                v[1] = *(etop+2);
                v[2] = *(etop+3);
                x = *(etop+4);
                noiz[0]=*(etop+5);
                noiz[1]=*(etop+6);
                noiz[2]=*(etop+7);
            }
            etop+=noiz_dim*2;
            pnoise3(noiz_dim, vec_result, noiz, x, v, y );
            *(--etop) = vec_result[Z];
            *(--etop) = vec_result[Y];
            *(--etop) = vec_result[X];
            cia+=2;
            break;
        case oAMBIENT:
            etop -= 3;
            ambient(etop, pi);
            break;
        case oDIFFUSE:
            diffuse(&(dataseg[P_DA_OFFS]), etop, pi, v);
            etop += 3;
            *(--etop) = v[2];
            *(--etop) = v[1];
            *(--etop) = v[0];
            break;
        case oPHONG:
            phong(&(dataseg[P_DA_OFFS]), etop+4, etop+1, *(etop), pi, v);
            etop += 7;
            *(--etop) = v[2];
            *(--etop) = v[1];
            *(--etop) = v[0];
            break;
        case oSPECULAR:
            specular(&(dataseg[P_DA_OFFS]), etop+4, etop+1, *(etop), pi, v);
            etop += 7;
            *(--etop) = v[2];
            *(--etop) = v[1];
            *(--etop) = v[0];
            break;
        case oSPECULARBRDF:
            pf = etop+7;
            VNormalizeEq(pf);
            pf = etop+4;
            VNormalizeEq(pf);
            pf = etop+1;
            VNormalizeEq(pf);
            specularbrdf(etop+7, etop+4, etop+1, *etop, v);
            etop += 10;
            *(--etop) = v[2];
            *(--etop) = v[1];
            *(--etop) = v[0];
            break;
        case oFACEFORWARD:
            f = *(etop  ) * *(etop+3) +
                *(etop+1) * *(etop+4) +
                *(etop+2) * *(etop+5);
            if (f > 0) {
                *(etop+6) = - *(etop+6);
                *(etop+7) = - *(etop+7);
                *(etop+8) = - *(etop+8);
            }
            etop += 6;
            break;
        case oDEPTH:
            x = dataseg[E_DA_OFFS  ] - *(etop  );
            y = dataseg[E_DA_OFFS+1] - *(etop+1);
            z = dataseg[E_DA_OFFS+2] - *(etop+2);
            f = sqrt(x*x + y*y + z*z);
            f /= Max_Distance;
            etop += 2;
            *etop = f;
            break;
        case oROTATE:
            rotate(v, etop+7, *(etop+6), etop+3, etop);
            etop += 10;
            *(--etop) = v[2];
            *(--etop) = v[1];
            *(--etop) = v[0];
            break;
        case oPTLINED:
            x = ptlen(etop+6, etop+3, etop);
            etop += 9;
            *(--etop) = x;
            break;
        case oPRINTF:
            i = GETSLOT;
            j = GETSLOT2;
            s_prf(pi->inst_strings[i].string, j, etop);
            cia += 4;
            etop += j;  /* count of args */
            break;
        case oEQ3:
            cond =  *(etop  ) == *(etop+3) &&
                *(etop+1) == *(etop+4) &&
                *(etop+2) == *(etop+5);
            etop += 6;
            if (cond)
                cia = GETSLOT;
            else
                cia += 2;
            break;
        case oEQ:
            y = *etop++;
            x = *etop++;
            if (x == y)
                cia = GETSLOT;
            else
                cia += 2;
            break;
        case oEQS:
            i = *etop++;
            j = *etop++;
            if (strcmp(pi->inst_strings[i].string, pi->inst_strings[j].string)==0)
                cia = GETSLOT;
            else
                cia += 2;
            break;
        case oNE3:
            cond =  *(etop  ) != *(etop+3) ||
                *(etop+1) != *(etop+4) ||
                *(etop+2) != *(etop+5);
            etop += 6;
            if (cond)
                cia = GETSLOT;
            else
                cia += 2;
            break;
        case oNE:
            y = *etop++;
            x = *etop++;
            if (x != y)
                cia = GETSLOT;
            else
                cia += 2;
            break;
        case oNES:
            i = *etop++;
            j = *etop++;
            if (strcmp(pi->inst_strings[i].string, pi->inst_strings[j].string))
                cia = GETSLOT;
            else
                cia += 2;
            break;
        case oLT:
            y = *etop++;
            x = *etop++;
            if (x < y)
                cia = GETSLOT;
            else
                cia += 2;
            break;
        case oGT:
            y = *etop++;
            x = *etop++;
            if (x > y)
                cia = GETSLOT;
            else
                cia += 2;
            break;
        case oLE:
            y = *etop++;
            x = *etop++;
            if (x <= y)
                cia = GETSLOT;
            else
                cia += 2;
            break;
        case oGE:
            y = *etop++;
            x = *etop++;
            if (x >= y)
                cia = GETSLOT;
            else
                cia += 2;
            break;
        case oJMP:
            cia = GETSLOT;
            break;
        case oSPLINE:
            j = GETSLOT; /* spline type */
            if (j == 0){
                i = *(etop++);
                j = spline_type_conv(pi, i);
            }
            x = *etop++; /* parameter */
            i = (int)*etop++; /* cpoint count */
            cia += 2;
            z = spline_f(x, i, j, etop);
            etop += i;
            *(--etop) = z;
            break;
        case oSPLINE3:
            j = GETSLOT; /* spline type */
            if (j == 0){
                i = *(etop++);
                j = spline_type_conv(pi, i);
            }
            x = *etop++; /* parameter */
            i = (int)*etop++; /* cpoint count */
            cia += 2;
            spline_p(v, x, i, j, etop);
            etop += 3*i;
            *(--etop) = v[2];
            *(--etop) = v[1];
            *(--etop) = v[0];
            break;
        case oTRANSFORM:
            {
                int to_space = GETSLOT;
                int from_space = GETSLOT2;
                cia+=4;
                if (from_space == 0){
                    i = *(etop++);
                    from_space = point_space_conv(pi, i);
                }
                if (to_space == 0){
                    i = *(etop++);
                    to_space = point_space_conv(pi, i);
                }
                v[X] = *(etop  );
                v[Y] = *(etop+1);
                v[Z] = *(etop+2);
                shader_transform(pi, vec_result, v, to_space, from_space, POINT_TRANSFORM);
                *(etop  ) = vec_result[X];
                *(etop+1) = vec_result[Y];
                *(etop+2) = vec_result[Z];
            }
            break;
        case oNTRANSFORM:
            {
                int to_space = GETSLOT;
                int from_space = GETSLOT2;
                cia+=4;
                if (from_space == 0){
                    i = *(etop++);
                    from_space = point_space_conv(pi, i);
                }
                if (to_space == 0){
                    i = *(etop++);
                    to_space = point_space_conv(pi, i);
                }
                v[X] = *(etop  );
                v[Y] = *(etop+1);
                v[Z] = *(etop+2);
                shader_transform(pi, vec_result, v, to_space, from_space, NORMAL_TRANSFORM);
                *(etop  ) = vec_result[X];
                *(etop+1) = vec_result[Y];
                *(etop+2) = vec_result[Z];
            }
            break;
        case oVTRANSFORM:
            {
                int to_space = GETSLOT;
                int from_space = GETSLOT2;
                cia+=4;
                if (from_space == 0){
                    i = *(etop++);
                    from_space = point_space_conv(pi, i);
                }
                if (to_space == 0){
                    i = *(etop++);
                    to_space = point_space_conv(pi, i);
                }
                v[X] = *(etop  );
                v[Y] = *(etop+1);
                v[Z] = *(etop+2);
                shader_transform(pi, vec_result, v, to_space, from_space, VECTOR_TRANSFORM);
                *(etop  ) = vec_result[X];
                *(etop+1) = vec_result[Y];
                *(etop+2) = vec_result[Z];
            }
            break;
        case oPOP:
            etop++;
            break;
        case oPOP3:
            etop+=3;
            break;
        case oPI:
            --etop;
            *etop = M_PI;
            break;
        case oREFLECT:
            reflect(v, etop+3, etop);
            etop += 3;
            *etop = v[0];
            *(etop+1) = v[1];
            *(etop+2) = v[2];
            break;
        case oREFRACT:
            refract(v, etop+4, etop+1, *etop);
            etop += 4;
            *etop = v[0];
            *(etop+1) = v[1];
            *(etop+2) = v[2];
            break;
        case oFRESNEL:
            {
                v[X] = *(etop+4);
                v[Y] = *(etop+5);
                v[Z] = *(etop+6);
                fresnel(v, etop+1, *etop, etop+5, etop+6);
                etop += 5;
            }
            break;
        case oFRESNEL2:
            {
                DBL eta = *(etop);
                DBL normal[3];
                v[X] = *(etop+4);
                v[Y] = *(etop+5);
                v[Z] = *(etop+6);
                normal[X] = *(etop + 1);
                normal[Y] = *(etop + 2);
                normal[Z] = *(etop + 3);
                fresnel(v, normal, eta, etop-1, etop);
                refract(etop+4, v, normal, eta);
                reflect(etop+1, v, normal);
            }
            etop -=1;
            break;
        case oCTRANSFORM:
            {
                int to_space = GETSLOT;
                int from_space = GETSLOT2;
                RGB colour;
                cia+=4;
                if (from_space == 0){
                    i = *(etop++);
                    from_space = color_space_conv(pi, i);
                }
                if (to_space == 0){
                    i = *(etop++);
                    to_space = color_space_conv(pi, i);
                }
                colour[0] = *etop;
                colour[1] = *(etop+1);
                colour[2] = *(etop+2);
                Transform_Colour(colour, to_space, from_space);
                *etop = colour[0];
                *(etop+1) = colour[1];
                *(etop+2) = colour[2];
            }
            break;
        case oTRACE:
            shader_trace(pi, v, etop+3, etop);
            etop += 6;
            *(--etop) = v[2];
            *(--etop) = v[1];
            *(--etop) = v[0];
            break;
        case oILLUM1:
            {
                int result;
                int illum_address = GETSLOT;
                COLOUR ClVal;
                f = M_PI;
                vec_result[X] = vec_result[Y] = vec_result[Z] = 1;
                v[0] = *(etop);
                v[1] = *(etop+1);
                v[2] = *(etop+2);
                if (in_illuminance() && 0){
                    /*  
                        avoid recursive calls during shadow calculations
                        from do_illuminance 
                    */ 
                    result = 0;
                }
                else{
                    result = do_illuminance(v, 
                            &(pi->Light), ClVal, vec_result, f, pi->Eye_Ray);
                }
                if (result){
                    dataseg[Cl_DA_OFFS] = ClVal[0]; 
                    dataseg[Cl_DA_OFFS+1] = ClVal[1]; 
                    dataseg[Cl_DA_OFFS+2] = ClVal[2]; 
                    dataseg[L_DA_OFFS] = pi->Light->Center[0] - 
                            dataseg[P_DA_OFFS];
                    dataseg[L_DA_OFFS+1] = pi->Light->Center[1]-
                            dataseg[P_DA_OFFS+1];
                    dataseg[L_DA_OFFS+2] = pi->Light->Center[2]-
                            dataseg[P_DA_OFFS+2];
                    cia = illum_address;
                }
                else{
                    pi->Light = NULL;
                    cia += 2;
                    etop += 3;
                }
            }
            break;
        case oILLUM3:
            {
                int result;
                int illum_address = GETSLOT;
                COLOUR ClVal;
                f = *(etop);
                        
                vec_result[X] = *(etop+1);
                vec_result[Y] = *(etop+2);
                vec_result[Z] = *(etop+3);
                v[0] = *(etop+4);
                v[1] = *(etop+5);
                v[2] = *(etop+6);
                result = do_illuminance(v, 
                        &(pi->Light), ClVal, vec_result, f, pi->Eye_Ray);
                if (result){
                    dataseg[Cl_DA_OFFS] = ClVal[0]; 
                    dataseg[Cl_DA_OFFS+1] = ClVal[1]; 
                    dataseg[Cl_DA_OFFS+2] = ClVal[2]; 
                    dataseg[L_DA_OFFS] = pi->Light->Center[0] - 
                            dataseg[P_DA_OFFS];
                    dataseg[L_DA_OFFS+1] = pi->Light->Center[1]-
                            dataseg[P_DA_OFFS+1];
                    dataseg[L_DA_OFFS+2] = pi->Light->Center[2]-
                            dataseg[P_DA_OFFS+2];
                    cia = illum_address;
                }
                else{
                    pi->Light = NULL;
                    cia += 2;
                    etop += 7;
                }
            }
            break;
        case oCALCULATENORMAL:
            break;
        case oCHK:
            if (etop != &e_stack[MAXSTACK]){
                Error("shader data stack synchronization error by %d bytes @%d!", 
                    etop-(&e_stack[MAXSTACK]) , cia-1);
            }
            break;
        default:
            Error("Wrong shader operand %d near address %d\n", op, cia);
        }     
#if (TEST_BOUNDS)
        if (etop > &e_stack[MAXSTACK]){
            /* Yes, messages are correct! Stack grows downward! */
            Error("shader data stack underflow!");
        }
        if (etop <  &e_stack[0]){
            Error("shader data stack overflow!");
        }
#endif        
    }
#if (TEST_BOUNDS)
    if (etop != &e_stack[MAXSTACK]){
        Error("shader data stack synchronization error by %d bytes!", 
            etop-(&e_stack[MAXSTACK]) );
    }
#endif    
}


int color_space_conv(shader_inst_t *pi, int offset)
{
    char *color_space =  pi->inst_strings[offset].string;
    
    if (strcmp(color_space, "rgb") == 0){
        return CS_RGB;
    }
    else if (strcmp(color_space, "hsv") == 0){
        return CS_HSV;
    }
    else if (strcmp(color_space, "hsl") == 0){
        return CS_HSL;
    }
    else if (strcmp(color_space, "xyz") == 0){
        return CS_XYZ;
    }
    else if (strcmp(color_space, "XYZ") == 0){
        return CS_XYZ;
    }
    else if (strcmp(color_space, "xyY") == 0){
        return CS_XY_Y;
    }
    else if (strcmp(color_space, "YIQ") == 0){
        return CS_YIQ;
    }
    else{
        Error("Incorrect color space in shader");
    }
}

int point_space_conv(shader_inst_t *pi, int offset)
{
    char *point_space =  pi->inst_strings[offset].string;
    if (strcmp(point_space, "raster") == 0){
        return PS_RASTER;
    }
    else if (strcmp(point_space, "screen") == 0){
        return PS_SCREEN;
    }
    else if (strcmp(point_space, "camera") == 0){
        return PS_CAMERA;
    }
    else if (strcmp(point_space, "world") == 0){
        return PS_WORLD;
    }
    else if (strcmp(point_space, "object") == 0){
        return PS_OBJECT;
    }
    else if (strcmp(point_space, "shader") == 0){
        return PS_SHADER;
    }
    else if (strcmp(point_space, "current") == 0){
        return PS_CURRENT;
    }
    else if (strcmp(point_space, "NDC") == 0){
        return PS_NDC;
    }
    else{
        Error("Incorrect point space in shader");
    }
}

int spline_type_conv(shader_inst_t *pi, int offset)
{
    char * spline_type = pi->inst_strings[offset].string;
    if (  strcmp(spline_type, "catmull_rom") == 0){
        return SP_CATMULL_ROM;
    }
    else if (  strcmp(spline_type, "bezier") == 0){
        return SP_BEZIER;
    }
    else if (  strcmp(spline_type, "bspline") == 0){
        return SP_BSPLINE;
    }
    else if (  strcmp(spline_type, "hermite") == 0){
        return SP_HERMITE;
    }
    else if (  strcmp(spline_type, "linear") == 0){
        return SP_LINEAR;
    }
    else{
        Error("Incorrect spline type in shader.");
    }
}

void reflect( DBL result[3],  DBL I[3], DBL N[3])
{
    DBL dot;
    
    VDot(dot, I, N);
    dot *= 2;
    result[X] = I[X] - dot*N[X];
    result[Y] = I[Y] - dot*N[Y];
    result[Z] = I[Z] - dot*N[Z];
}


void refract(DBL result[3], DBL I[3], DBL N[3], DBL eta)
{

    DBL dot = I[X]*N[X] +  I[Y]*N[Y] + I[Z]*N[Z];
    DBL k = 1 - eta*eta*(1 - dot*dot);
    if (k<0.0){
        result[X]=result[Y]=result[Z]=0.0;
    }
    else{
        DBL coeff = eta*dot + sqrt(k);
        result[X]= eta*I[X] - coeff * N[X];
        result[Y]= eta*I[Y] - coeff * N[Y];
        result[Z]= eta*I[Z] - coeff * N[Z];
    }
}

void fresnel(DBL I[3], DBL N[3], DBL eta, DBL *pkr, DBL *pkt)
{
    DBL dot, refr_dot;
    DBL sum, diff;
    DBL I_N_angle, R_N_angle;
    DBL t, t1;
    
    if (eta > 1.0){
        *pkr = 1;
        *pkt = 0;
        return;
    }
    VNormalizeEq(N);
    VNormalizeEq(I);
    
    VDot(dot, I, N);
    
    if (dot <0){
        dot = -dot;
    }
    I_N_angle = acos(dot);
    R_N_angle = asin(sin(I_N_angle)*eta);
    sum = I_N_angle + R_N_angle;
    diff = I_N_angle - R_N_angle;
    t = sin(diff);
    t1 = t*t;
    t = sin(sum);
    t1/=(t*t);
    *pkr = t1;
    t = tan(diff);
    t1 = t*t;
    t = tan(sum);
    t1 /= (t*t);
    *pkr += t1;
    *pkr /= 2.0;
    *pkt = 1 - *pkr;
}



void ds(DBL *etop, DBL e_stack[])
{
    DBL *e;

    if (etop < &e_stack[MAXSTACK]) {
        for(e = &e_stack[MAXSTACK-1]; e >= etop; e--)
            if (e == etop)
                printf("    --->%lf\n", *e);
            else
                printf("\t%lf\n", *e);
    }
}

void s_prf(char *fmt, int nargs, DBL *etop)
{
    char outbuf[512], *p, *q;
    int ai = 0;
    p = outbuf;
    while (*fmt != '\0') {
        if (*fmt == '\\') {
            switch (*++fmt) {
            case 'n': *p++ = '\n'; break;
            case 'r': *p++ = '\r'; break;
            case 't': *p++ = '\t'; break;
            case 'f': *p++ = '\f'; break;
            default:
                *p++ = *fmt;
            }
        } else if (*fmt == '%') {
            switch (*++fmt) {
            case 'p':
                p += sprintf(p, "%g %g %g",
                    *(etop+ai),
                    *(etop+ai+1),
                    *(etop+ai+2));
                ai += 3;
                break;
            case 'f':
                p += sprintf(p, "%g", *(etop+ai));
                ai += 1;
                break;
            default:
                *p++ = *fmt;
            }
        } else {
            *p++ = *fmt;
        }
        fmt++;
    }
    *p = '\0';
    Debug_Info(outbuf);
}

void phong(DBL *p, DBL *n, DBL *e, DBL size, shader_inst_t *pi, DBL *c)
{
    DBL cone = M_PI_2;
    LIGHT_SOURCE *light = NULL;
    COLOUR ClVal;
    DBL refl[3];
    DBL dot;
    DBL light_vec[3];
    
    DBL coeff;
    
    c[0] = 0.;
    c[1] = 0.;
    c[2] = 0.;
    if (in_illuminance()){
        return;
    }
    VNormalizeEq(n);
    VNormalizeEq(e);
    VScaleEq(e, -1);
    reflect(refl, e, n);
    
    while (do_illuminance(p, &light, ClVal, n, cone, pi->Eye_Ray) != 0)
    {

        VSub(light_vec, light->Center, p);
        VNormalizeEq(light_vec);
        VDot(dot, refl, light_vec);
        
        if (dot >0.0){
            coeff = pow(dot, size);
            c[0] += ClVal[0]*coeff;
            c[1] += ClVal[1]*coeff;
            c[2] += ClVal[2]*coeff;
        }
    }
}

void specular(DBL *p, DBL *n, DBL *e, DBL roughness, shader_inst_t *pi, DBL *c)
{
    DBL cone = M_PI_2;
    LIGHT_SOURCE *light = NULL;
    COLOUR ClVal;
    DBL light_vec[3];
    DBL c1[3];
    
    c[0] = 0.;
    c[1] = 0.;
    c[2] = 0.;
    
    if (in_illuminance()){
        return;
    }
    VNormalizeEq(n);
    VNormalizeEq(e);

    while (do_illuminance(p, &light, ClVal, n, cone, pi->Eye_Ray) != 0)
    {

        VSub(light_vec, light->Center, p);
        VNormalizeEq(light_vec);
        specularbrdf(light_vec, n, e, roughness, c1);
        c[0] += ClVal[0]*c1[0];
        c[1] += ClVal[1]*c1[1];
        c[2] += ClVal[2]*c1[2];        
    }
}

/* assume normalized vectors! */
void specularbrdf(DBL *p, DBL *n, DBL *e, DBL roughness, DBL *c)
{
    DBL h[3];
    DBL dot;
    VAdd(h, p, e);
    VNormalizeEq(h);
    VDot(dot, n, h);

    if (dot >0.0){
        DBL coeff = pow(dot, 1/roughness);
        c[0] = c[1] = c[2] = coeff;
    }
    else{
        c[0] = c[1] = c[2] = 0.0;
    }
}


void diffuse(DBL *p, DBL *n, shader_inst_t *pi, DBL *c)
{

    DBL cone = M_PI_2;
    LIGHT_SOURCE *light = NULL;
    COLOUR ClVal;
    DBL dot;
    DBL light_vec[3];
    
    c[0] = 0.;
    c[1] = 0.;
    c[2] = 0.;
    
    if (in_illuminance()){
        return;
    }

    VNormalizeEq(n);

    while (do_illuminance(p, &light, ClVal, n, cone, pi->Eye_Ray) != 0)
    {

        VSub(light_vec, light->Center, p);
        VNormalizeEq(light_vec);
        VDot(dot, n, light_vec);
        if (dot >0.0){
            c[0] += ClVal[0]*dot;
            c[1] += ClVal[1]*dot;
            c[2] += ClVal[2]*dot;
        }
    }
}

void ambient(DBL *c, shader_inst_t * pi)
{
    if (pi->ambient_init){
        c[0] = pi->data[As_DA_OFFS  ];
        c[1] = pi->data[As_DA_OFFS+1];
        c[2] = pi->data[As_DA_OFFS+2];
    }
    else{
        c[0] = 0.0;
        c[1] = 0.0;
        c[2] = 0.0;
    }
}

void hermite_basis(DBL u, DBL bn[4])
{
    DBL usq = u*u, u2 = u+u;
    bn[0] = usq*(u2 - 3.0) + 1.0;
    bn[1] = u*(usq - u2 + 1.0);
    bn[2] = usq*(3.0 - u2);
    bn[3] = usq*(u - 1.0);
}

void cat_rom_basis(DBL u, DBL bn [4])
{

    DBL usq = u*u;
    bn[0] = (usq*(-u + 2.0) - u)/2.0;
    bn[1] = (usq*(3.0*u - 5.0)+2.0)/2.0;
    bn[2] = (usq*(-3.0*u + 4.0) +u)/2.0;
    bn[3] = usq*(u - 1.0)/2.0;
}

void bezier_basis(DBL u, DBL bn [4])
{

    DBL diff = 1.0 - u;
    bn[0] = diff*diff*diff;
    bn[1] = 3.0*u*diff*diff;
    bn[2] = 3.0*u*u*diff;
    bn[3] = u*u*u;
}

void bspline_basis(DBL u, DBL bn [4])
{
    /*
        Note: don't remove those constants from this function: although
        it would be more efficient to divide by 6 only once, (after
        multiplying with control points), but such approach accumulates error
        faster than desirable...
    */
    DBL diff = 1.0 - u;
    DBL usq = u*u;
    bn[0] = (diff*diff*diff)/6.0;
    bn[1] = (usq*(0.5*u-1.0)) +2.0/3.0;
    bn[2] = (usq*diff + u)/2.0 +1.0/6.0;
    bn[3] = (usq*u)/6.0;
}


#define pi(n)   (etop + 3*(n))


void ptangent(DBL t[3], int i, int npoints, DBL *etop)
{
    DBL u[3], v[3];
    if (i == 0){
        VSub(t, pi(i+1), pi(i));
    }
    else if (i == npoints-1){
        VSub(t, pi(i), pi(i-1));
    }
    else {
        VSub(u, pi(i+1), pi(i));
        VSub(v, pi(i), pi(i-1));
        t[0] = (u[0] + v[0])/2.;
        t[1] = (u[1] + v[1])/2.;
        t[2] = (u[2] + v[2])/2.;
    }
}

void splineval_point(int type, DBL u, DBL p[4][3], DBL c[3])
{
    DBL b[4];
    switch (type){
        case SP_HERMITE:
            hermite_basis(u, b);
            break;
        case SP_BEZIER:
            bezier_basis(u, b);
            break;
        case SP_CATMULL_ROM:
            cat_rom_basis(u, b);
            break;
        case SP_BSPLINE:
            bspline_basis(u, b);
            break;
    }
    c[0] = p[0][0]*b[0] + p[1][0]*b[1] + p[2][0]*b[2] + p[3][0]*b[3];
    c[1] = p[0][1]*b[0] + p[1][1]*b[1] + p[2][1]*b[2] + p[3][1]*b[3];
    c[2] = p[0][2]*b[0] + p[1][2]*b[1] + p[2][2]*b[2] + p[3][2]*b[3];
}


void spline_p(DBL *c, DBL u, int npoints, int type, DBL * etop)
{
    DBL t, frac;
    DBL hv[4][3];
    int i;

    if (u <= 0.0) {
        c[0] = pi(1)[0];
        c[1] = pi(1)[1];
        c[2] = pi(1)[2];
        return;
    }
    if (u >= 1.0) {
        c[0] = pi(npoints-2)[0];
        c[1] = pi(npoints-2)[1];
        c[2] = pi(npoints-2)[2];
        return;
    }

    t = u*(npoints-3)+1;
    i = (int)t;
    frac = t -((DBL)i);
    if (type == SP_LINEAR){
        c[0] = frac*(pi(i+1)[0] - pi(i)[0])+ pi(i)[0];
        c[1] = frac*(pi(i+1)[1] - pi(i)[1])+ pi(i)[1];
        c[2] = frac*(pi(i+1)[2] - pi(i)[2])+ pi(i)[2];
        return;
    }
    else if (type ==SP_HERMITE){
        hv[0][0] = pi(i)[0]; hv[0][1] = pi(i)[1]; hv[0][2] = pi(i)[2];
        ptangent(hv[1], i, npoints, etop);
        hv[2][0] = pi(i+1)[0]; hv[2][1] = pi(i+1)[1]; hv[2][2] = pi(i+1)[2];
        ptangent(hv[3], i+1, npoints, etop);
    }    
    else{
        hv[0][0] = pi(i-1)[0]; hv[0][1] = pi(i-1)[1]; hv[0][2] = pi(i-1)[2];
        hv[1][0] = pi(i  )[0]; hv[1][1] = pi(i  )[1]; hv[1][2] = pi(i  )[2];
        hv[2][0] = pi(i+1)[0]; hv[2][1] = pi(i+1)[1]; hv[2][2] = pi(i+1)[2];
        hv[3][0] = pi(i+2)[0]; hv[3][1] = pi(i+2)[1]; hv[3][2] = pi(i+2)[2];
    
    }
    splineval_point(type, frac, hv, c);
}

#define fi(n)   (etop + (n))

DBL ftangent(int i, int npoints, DBL *etop)
{
    DBL u, v;
    if (i == 0)
        return *fi(i+1) - *fi(i);
    else if (i == npoints-1)
        return *fi(i) - *fi(i-1);
    else {
        u = *fi(i+1) - *fi(i);
        v = *fi(i) - *fi(i-1);
        return (u+v)/2.;
    }
}

DBL splineval_float(int type, DBL u, DBL p[4])
{
    DBL b[4];
    switch (type){
        case SP_HERMITE:
            hermite_basis(u, b);
            break;
        case SP_BEZIER:
            bezier_basis(u, b);
            break;
        case SP_CATMULL_ROM:
            cat_rom_basis(u, b);
            break;
        case SP_BSPLINE:
            bspline_basis(u, b);
            break;
    }

    return p[0]*b[0] + p[1]*b[1] + p[2]*b[2] + p[3]*b[3];
}


DBL spline_f(DBL u, int npoints, int type, DBL *etop)
{
    DBL t;
    DBL hv[4];
    int i;
    if (type != SP_BSPLINE){
        if (u <= 0.0)
            return *fi(1);
        if (u >= 1.0)
            return *fi(npoints-2);
    }

    t = u*(npoints-3)+1;
    i = (int)t;
    if (type == SP_LINEAR){
        return fmod(t,1)*(*fi(i+1) - *fi(i))+ *fi(i);
    }
    else if (type ==SP_HERMITE){
        hv[0] = *fi(i);
        hv[1] = ftangent(i, npoints, etop);
        hv[2] = *fi(i+1);
        hv[3] = ftangent(i+1, npoints, etop);
    }
    else{
        hv[0] = *fi(i-1);
        hv[1] = *fi(i);
        hv[2] = *fi(i+1);
        hv[3] = *fi(i+2);
    }
    return splineval_float(type, fmod(t,1), hv);
        
}

DBL cellnoise(int dim, DBL vector[3], DBL time)
{
    int temp, seed;
    DBL result;
    if (dim==1)
        seed = Hash1d((int) vector[X], 42);
    else if (dim==2)
        seed = Hash2d((int)vector[X], (int)vector[Y]);
    else
        seed = Hash3d((int)vector[X], (int)vector[Y], (int)vector[Z]);
    if (dim == 4)
        seed = Hash2d(seed, (int)time);
    temp = POV_GET_OLD_RAND();
    POV_SRAND(seed);
    result = FRAND();
    POV_SRAND(temp);
    return min(result, 1.0);
}


void cellnoise3d(int dim, DBL result[3], DBL vector[3], DBL time)
{

    int temp, seed;
    
    if (dim==1)
        seed = Hash1d((int) vector[X], 42);
    else if (dim==2)
        seed = Hash2d((int)vector[X], (int)vector[Y]);
    else
        seed = Hash3d((int)vector[X], (int)vector[Y], (int)vector[Z]);
    if (dim == 4)
        seed = Hash2d(seed, (int)time);

    temp = POV_GET_OLD_RAND();
    POV_SRAND(seed);
    result[X] = FRAND();
    result[Y] = FRAND();
    result[Z] = FRAND();
    POV_SRAND(temp);
    
    if (result[X]>1.0) result[X] = 1.0;
    if (result[Y]>1.0) result[Y] = 1.0;
    if (result[Z]>1.0) result[Z] = 1.0;
}

DBL pnoise1(int dim, DBL point[3], DBL time, 
                        DBL pperiod[3], DBL tperiod)
{
    /* NOTE: does not support 4-dim noise ! */
    long period_val;
    DBL tempp[3];
    DBL ret_val;
    
    period_val = ((long)point[0])/((long)pperiod[0]);
    if (point[0] < 0.0){
        /* shift value to get later to value in period above 0 */
        --period_val;
    }
    tempp[0] = point[0] - period_val * pperiod[0];
    if (dim == 1){
        tempp[1] = tempp [2] = tempp[0];
    }
    else{
        period_val = ((long)point[2])/((long)pperiod[2]);
        if (point[2] < 0.0){
            --period_val;
        }
        tempp[2] = point[2] - period_val * pperiod[2];
        if (dim == 2){
            tempp[1] = tempp[0];
        }
        else{
            period_val= ((long)point[1])/((long)pperiod[1]);
            if (point[1] < 0.0){
                --period_val;
            }
            tempp[1] = point[1] - period_val * pperiod[1];    
        }
    }
    ret_val = Noise(tempp);
    return ret_val;
}


void pnoise3(int dim, DBL resultp[3], 
                      DBL point[3], DBL time,
                      DBL pperiod[3], DBL tperiod)
{

    /* NOTE: does not support 4-dim noise ! */
    long period_val;
    DBL tempp[3];
    
    period_val = ((long)point[0])/((long)pperiod[0]);
    if (point[0] < 0.0){
        --period_val;
    }
    tempp[0] = point[0] - period_val * pperiod[0];
    if (dim == 1){
        tempp[1] = tempp [2] = tempp[0];
    }
    else{
        period_val = ((long)point[2])/((long)pperiod[2]);
        if (point[2] < 0.0){
            --period_val;
        }
        tempp[2] = point[2] - period_val * pperiod[2];
        if (dim == 2){
            tempp[1] = tempp[0];
        }
        else{
            period_val= ((long)point[1])/((long)pperiod[1]);
            if (point[1] < 0.0){
                --period_val;
            }
            tempp[1] = point[1] - period_val * pperiod[1];    
        }
    }
    DNoise(resultp, tempp);


}
/* 
    Rotation matrix  is taken from _the_ "Computer Graphics 2nd ed. in C"
    ex. 5.15, pg 227 with optimisation in common subexpression calculation.
*/

void rotate(DBL resultp[3], DBL point[3], DBL angle, DBL startv[3],
        DBL endv[3])
{

    DBL anglecos = cos(angle);
    DBL anglesin = sin(angle);
    DBL vector[3];
    DBL p[3];
    DBL tmp, tmp1;
    
    VSub(vector, endv, startv);
    VSub(p, point, startv);
    
    tmp = vector[0]*vector[0];
    resultp[0] = p[0] * (tmp + anglecos*(1-tmp));
    tmp = vector[1]*vector[1];
    resultp[1] = p[1] * (tmp + anglecos*(1-tmp));
    tmp = vector[2]*vector[2];
    resultp[2] = p[2] * (tmp + anglecos*(1-tmp));
    
    tmp = vector[0]*vector[1]*(1-anglecos);
    tmp1 = vector[2]*anglesin;
    resultp[0] += p[1]*(tmp-tmp1);
    resultp[1] += p[0]*(tmp+tmp1);
    
    tmp = vector[2]*vector[0]*(1-anglecos);
    tmp1 = vector[1]*anglesin;
    resultp[0] += p[2]*(tmp+tmp1);
    resultp[2] += p[0]*(tmp-tmp1);
    
    tmp = vector[1]*vector[2]*(1-anglecos);
    tmp1 =  vector[0]*anglesin;
    resultp[1] += p[2]*(tmp-tmp1);
    resultp[2] += p[1]*(tmp+tmp1);

    VAddEq(resultp, startv);    
}

DBL ptlen(DBL p1[3], DBL p2[3], DBL pq[3])
{
    DBL linevec[3];
    DBL pvec[3];
    DBL crossv[3];
    DBL dot, dot2;
    
    VSub(linevec, p2, p1);
    VSub(pvec, pq, p1);
    VDot(dot, linevec, pvec);
    if (dot < 0){
        /* point is not on line, nearest is p1 */
        VLength(dot, pvec);
        return dot;
    }
    VDot(dot2, linevec, linevec);
    if (dot > dot2){
        /* not on line, p2 is nearest */
        VSub(pvec, pq, p2);
        VLength(dot, pvec);
        return dot;
    }
    /* point is on line */
    VCross(crossv, linevec, pvec);
    VLength(dot, crossv);
    VLength(dot2, linevec);
    dot /= dot2;
    return dot;
}

void shader_trace(
        shader_inst_t * pi, 
        DBL trace_col[3], 
        DBL startp[3], 
        DBL direction[3])
{

    COLOUR POVCol;
    RAY POVRay;
    POVRay.Initial[0]=startp[0];
    POVRay.Initial[1]=startp[1];
    POVRay.Initial[2]=startp[2];
    
    POVRay.Direction[0]=direction[0];
    POVRay.Direction[1]=direction[1];
    POVRay.Direction[2]=direction[2];
    
    VNormalize(POVRay.Direction, POVRay.Direction);
    
    Make_ColourA(POVCol, 0.0, 0.0, 0.0, 0.0, 0.0);
    Copy_Ray_Containers(&POVRay, pi->Eye_Ray);
    Trace_Level++;
    Trace( &POVRay, POVCol, 1.0, NULL);
    Trace_Level--;
    trace_col[0] = POVCol[0];
    trace_col[1] = POVCol[1];
    trace_col[2] = POVCol[2];
    
}



static int illuminance=0;
static int illuminance_recursion_depth = SHADER_RECURSION_DEPTH;
int in_illuminance()
{
    return illuminance;
}

void Set_Shader_Recursion(int new_depth)
{
    illuminance_recursion_depth = new_depth;
}

int do_illuminance(
    VECTOR Intersection, 
    LIGHT_SOURCE **Light,
    COLOUR Light_Colour,
    VECTOR Cone_Axis,
    DBL Cone_Angle,
    RAY *Eye_Ray)
{
    
    int Light_OK = FALSE;
    DBL Light_Source_Depth;
    RAY Light_Source_Ray;
    DBL Light_Angle_Cos;
    DBL Axis_Len;
    DBL Dir_Len;
    int Shadowed = FALSE;
    
    if (illuminance >= illuminance_recursion_depth){
        return 0;
    }
    
    while ( !Light_OK)
    {
        if (*Light == NULL)
            *Light = Frame.Light_Sources;
        else
            *Light = (*Light)->Next_Light_Source;
        if ((*Light) == NULL)
        {
            break;
        }
        illuminance += 1;

        Shadowed = Test_Shadow(*Light, &Light_Source_Depth, 
            &Light_Source_Ray, Eye_Ray, Intersection, Light_Colour);
        illuminance -= 1;

        /* determine, whether light source is in specified cone */
        VDot(Light_Angle_Cos, Cone_Axis, Light_Source_Ray.Direction);
        VLength(Axis_Len, Cone_Axis);
        VLength(Dir_Len, Light_Source_Ray.Direction);
        Light_Angle_Cos /= Axis_Len*Dir_Len;
        if (cos(Cone_Angle) > Light_Angle_Cos){
            /* light is outside of cone, skip it */
            continue;
        }
        /* Light source is found, return true to indicate this*/
        return 1;
    }
    return 0;
}

#endif /* ShaderPatch */
