/****************************************************************************

FILE	shdrglue.c

PURPOSE 
    	This file provides "gluing" functionality between shader and POV-Ray.
	
*****************************************************************************/
/*****************************************************************************                                                                              
 Copyright (c) 1995
      Robert J. Mercier
 Copyright 2000
      Vahur Krouverk

THIS CODE IS SUPPLIED ``AS IS'' WITH NO WARRANTY EXPRESSED
OR IMPLIED.  NO CLAIMS IS MADE TO IT'S SUITABILITY FOR
ANY PURPOSE AND NO RESPONSIBILITY IS ACCEPTED FOR ANY DAMAGES
OF ANY KIND RESULTING FROM IT'S USE.

When (if) this code is included into POV the usual POV legal stuff will apply.

*****************************************************************************/
#include "frame.h"
#ifdef ShaderPatch

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>

#include "frame.h"
#include "matrices.h"
#include "povray.h"
#include "vector.h"
#include "shdrvars.h"
#include "opcodes.h"
#include "mem.h"
#include "povproto.h"

static int ncalls=0;

#define cpy2shdr(A,B) \
    (pi->data[(A)]=(B)[X],pi->data[(A)+1]=(B)[Y],pi->data[(A)+2]=(B)[Z])
#define cpy2pov(A,B) \
    ((A)[X]=pi->data[(B)],(A)[Y]=pi->data[(B)+1],(A)[Z]=pi->data[(B)+2])


enum SHADOW_TYPE query_shadow_type(PIGMENT * Pigment)
{
    SHADER *shader = Pigment->Vals.Shader;

    shader_inst_t *templ_shader=(shader_inst_t*)shader;
    return templ_shader->shadow;
}


void Exec_Shader(
    PIGMENT *Pigment,
    INTERSECTION *Intersection,
    RAY * iray,
    VECTOR norm,
    COLOUR col,
    VECTOR i_point,
    int Shadow)
{
    SHADER *shader = Pigment->Vals.Shader;

    shader_inst_t *templ_shader=(shader_inst_t*)shader;
    shader_inst_t *pi=NULL;
    int data_size;
    int total_instance_size = sizeof(shader_inst_t);
    int string_size = 0;
    DBL norm_dot;    
    int i;
    shader_t *p=NULL;
    UV_VECT uv_vect;
    VECTOR RayVect;
    
    ncalls++;

    if (Shadow){
        if (templ_shader->shadow == No_Shadow){
            col[RED]=col[GREEN]=col[BLUE]=1.0;
            col[FILTER] = 0.0;
            col[TRANSM] = 1.0;
            return;
        }
        else if (templ_shader->shadow == Opaque_Shadow){
            col[RED]=col[GREEN]=col[BLUE]=1.0;
            col[FILTER] = 0.0;
            col[TRANSM] = 0.0;            
            return;
        }
        else if (templ_shader->shadow == Os_Shadow){
            col[RED]=col[GREEN]=col[BLUE]=1.0;
            col[FILTER] = 0.0;
            if (templ_shader->Os_arg != NULL){
                col[TRANSM] = 1.0 - templ_shader ->Os_arg->v.f;
            }
            else{
                col[TRANSM] = 0.0;
            }
            return;
        }
    }

    if (iray == NULL){
        static RAY unit_ray;
        static int init = 0;
        if (!init){
            init = 1; 
            unit_ray.Initial[X] = 0;
            unit_ray.Initial[Y] = 0;
            unit_ray.Initial[Z] = -1;
            unit_ray.Direction[X] = 0;
            unit_ray.Direction[Y] = 0;
            unit_ray.Direction[Z] = 1;
        }
        iray = &unit_ray;
    }

    p = templ_shader->shader;    
    
    /* 
        calculate total size for shader instance to allocate memory for instance
        structure and data once. 
        WARNING! This optimization may be nonportable, if data alignment
        for data (e.g. for DBL) is required or pointers for different data
        differ! If problems occur, then undef RIDE_FAST_SHADER
    */
#define RIDE_FAST_SHADER 1
    data_size = sizeof(DBL)*templ_shader->shader->dsize;
    if (p->dynastring){
        string_size = sizeof(stringtab)*(p->nstring);
    }
#if (RIDE_FAST_SHADER)
    total_instance_size += data_size + string_size;
#endif

    pi = (shader_inst_t *) POV_MALLOC(total_instance_size,
             "shader instance");
    memcpy(pi, templ_shader, sizeof(shader_inst_t));
    
#if (RIDE_FAST_SHADER)
    pi->data = (DBL*) (void*)(((char*)pi) + sizeof(shader_inst_t));
#else
    pi->data = (DBL*) POV_MALLOC(data_size, "shader instance data");
#endif
    memcpy(pi->data, p->data_templ, data_size);

    if (p->dynastring){
#if (RIDE_FAST_SHADER)    
        pi->inst_strings = (stringtab*)(void*)(((char*)(pi->data))+data_size);
#else
        pi->inst_strings = (stringtab*)POV_MALLOC(string_size, 
                "shader instance strings");
#endif
        for (i=0; i < p->nstring; i++){
            pi->inst_strings[i].allocated=0;
            pi->inst_strings[i].string=p->strings[i];
        }
    }

    if (Intersection){
        /* use intersection point from intersection structure: intersection
           point in additional parameter contains uv coordinates 
           of intersection, if surface uses uv mapping */
        cpy2shdr(P_DA_OFFS, Intersection->IPoint);
        /* calculate true view vector: iray->direction has normalized vector! */
        VSub(RayVect, Intersection->IPoint, iray->Initial);
        cpy2shdr(I_DA_OFFS, RayVect);
        cpy2shdr(Ng_DA_OFFS, Intersection->INormal);
        if (p->uvused){
            UVCoord(uv_vect, Intersection->Object, Intersection);    
            pi->data[u_DA_OFFS] = uv_vect[0];
            pi->data[v_DA_OFFS] = uv_vect[1];
            pi->data[s_DA_OFFS] = uv_vect[0];
            pi->data[t_DA_OFFS] = uv_vect[1];
        }
    
        if (norm){
            cpy2shdr(N_DA_OFFS, norm);
    
            VDot(norm_dot, norm, Intersection->INormal);
            if (norm_dot <0){
                /*
                  normal was flipped to point toward camera. Shader expects non-flipped
                  normals, so we flip it back (to geometry normal) here. 
                  On return we flip it again...
                */ 
                VScaleEq(&(pi->data[N_DA_OFFS]), -1.0);
            }
        }
    }
    else{
        /* try to find required values by other means */
        cpy2shdr(P_DA_OFFS, i_point);    
        /* issue warning about incomplete data */
        if (templ_shader->warning_flag == 0){
            Warning(0,  "Shader '%s' attached to non-object, initialization incomplete!\n", p->name);
            templ_shader->warning_flag = 1;
        }
    }
    
    
    pi->data[E_DA_OFFS]   = Frame.Camera->Location[X];
    pi->data[E_DA_OFFS+1] = Frame.Camera->Location[Y];
    pi->data[E_DA_OFFS+2] = Frame.Camera->Location[Z];

    pi->data[time_DA_OFFS] = opts.FrameSeq.Clock_Value;

    pi->Current_Pigment = Pigment;
    pi->Eye_Ray = iray;
    pi->Light = NULL;

    s_execute(pi); 
    
    /*  
        copy back instance startaddress from executed shader: this allows
        to handle each instance separately and avoid call for parameter copy
        if parameters for current instance are not passed from POVRay
    */
    templ_shader->inst_startaddr = pi->inst_startaddr;

    col[RED]   = pi->data[Ci_DA_OFFS  ];
    col[GREEN] = pi->data[Ci_DA_OFFS+1];
    col[BLUE]  = pi->data[Ci_DA_OFFS+2];
    /*
       NOTE: Possible problem:  povray uses only 1 channel for transmittance,
       so we just copy first channel. Usually opacity is same for each color 
       in shaders, so this shouldn't be problem.
    */
    col[TRANSM] = 1-pi->data[Oi_DA_OFFS];
    col[FILTER] = 0;

    if (norm){
        cpy2pov(norm, N_DA_OFFS);
        if (norm_dot<0){
            /* 
              parameter normal was flipped, flip it back to ensure correct
              lighting 
            */
            VScaleEq(norm, -1.0);
        }
    }
    
    if (p->dynastring){
        pi->inst_strings = (stringtab*)POV_MALLOC(sizeof(stringtab)*p->nstring,
                "shader instance string table");
        for (i=0; i < p->nstring; i++){
            if ( pi->inst_strings[i].allocated !=0){
                POV_FREE(pi->inst_strings[i].string);
            }
        }
#if !(RIDE_FAST_SHADER)
        POV_FREE(pi->inst_strings);
#endif        
    }

#if !(RIDE_FAST_SHADER)
    POV_FREE(pi->data);
#endif
    POV_FREE(pi);
}

double Exec_Iso_Shader(FUNCTION *func, VECTOR XYZ){

    shader_inst_t *pi=(shader_inst_t*)func->Shader;
    if (pi->data == NULL){
        int data_size = pi->shader->dsize * sizeof(DBL);
        pi->data = (DBL*) POV_MALLOC(data_size, "shader instance data vector");
        memcpy(pi->data, pi->shader->data_templ, data_size);
    }
    cpy2shdr(P_DA_OFFS, XYZ);
    ncalls++;
    s_execute(pi); 
    return pi->data[iso_val_DA_OFFS];

}

        
void transformation_matrix(
    shader_inst_t *Shader,                                                           
    int SpaceType, 
    TRANSFORM *Transform)
{
    MIdentity(Transform->matrix);
    
    switch(SpaceType){
        case PS_SHADER:
            {
    
                PIGMENT * Pigment = Shader->Current_Pigment;                                                                            
                WARP *warp=Pigment->Warps;                                                  
                while(warp != NULL){
                    if(warp->Warp_Type == TRANSFORM_WARP){
                        MTimes(Transform->matrix, Transform->matrix, (((TRANS*)warp)->Trans.matrix));
                    }
                    warp = warp->Next_Warp;
                }
            }
            break;
            
        case PS_OBJECT:
            /* 
                not implemented, as there is no single method for obtaining
                transformation mx for given object 
            */
            break;
        case PS_CAMERA:
            /*
                TODO
            */
            break;

        case PS_WORLD:
            /* transform is from/to world coordinates, ignore it */
            break;
        
        /* unimpelemented transformations */
        case PS_RASTER:
        case PS_SCREEN:
        case PS_NDC:
            break;
    }
}                                                                               

void shader_transform(                                                         
    shader_inst_t *Shader,                                                           
    VECTOR Result,                                                              
    VECTOR Initial,
    int To_Space,
    int From_Space,
    int Transform_Type)
{   
    
    TRANSFORM Trans;
    
    if (From_Space != PS_WORLD && To_Space != PS_WORLD){
        VECTOR Tmp;
        shader_transform(Shader, Tmp, Initial, PS_WORLD, From_Space,  Transform_Type);
        shader_transform(Shader, Result, Tmp, To_Space, PS_WORLD, Transform_Type);
        return;
    }
    else if (From_Space == PS_WORLD){
        transformation_matrix(Shader, To_Space, &Trans);
        MInvers(Trans.inverse, Trans.matrix);
        switch (Transform_Type){
            case POINT_TRANSFORM:
                MInvTransPoint(Result, Initial, &Trans);
                break;
            case NORMAL_TRANSFORM:
                MInvTransNormal(Result, Initial, &Trans);
                break;
            case VECTOR_TRANSFORM:
                MInvTransDirection(Result, Initial, &Trans);
                break;
        }
    }
    else if (To_Space == PS_WORLD){
        transformation_matrix(Shader, From_Space, &Trans);
        switch (Transform_Type){
            case POINT_TRANSFORM:
                MTransPoint(Result, Initial, &Trans);
                break;
            case NORMAL_TRANSFORM:
                MTransNormal(Result, Initial, &Trans);
                break;
            case VECTOR_TRANSFORM:
                MTransDirection(Result, Initial, &Trans);
                break;
        }
    }
}


void Initialize_Shader()
{
    int i;
    for (i=0; i< HSIZE; i++){
      /* clear shaders array */
      shaders[i] = NULL;
    }
    ncalls = 0;
    reset_shader_count();
}


int s_ncalls()
{
    return ncalls;
}


void Use_Shader(SHADER *shader)
{

    shader_inst_t *shdr=(shader_inst_t*)shader;
    shdr->refcount++;
}


SHADER * Copy_Shader(SHADER *shader){

    Use_Shader(shader);
    /*  shaders are refcounted and on each invocation separate copy is created,
        so we can just share instances 
    */
    return shader;
}


void Destroy_Shader(SHADER * shader)
{

    shader_inst_t *shdr= NULL;
    shader_t *p = NULL;
    pov_arg_t *args =NULL;

    if (shader == NULL){
        return;
    }
    
    shdr=(shader_inst_t*)shader;
    p = shdr->shader;
    args=shdr->args;
    
    shdr->refcount--;
    if (shdr->refcount > 0){
        return;
    }
    
    /* free argument list */    
    while(args != NULL){
        pov_arg_t *next = args->next;
        POV_FREE(args->name);
        POV_FREE(args);
        args = next;
    }
    
    /* free data vector */
    if (shdr->data != NULL){
        POV_FREE(shdr->data);
    }
    if (shdr->inst_strings != NULL){
        int i;
        for (i=0; i < p->nstring; i++){
            if (shdr->inst_strings[i].allocated){
                POV_FREE(shdr->inst_strings[i].string);
            }
        }
        POV_FREE(shdr->inst_strings);
    }

    p->refcount--;
    if (p->refcount == 0){
        int i;
        /* remove shader_t from shaders list */
        int slot = s_hash(p->filename);
        shader_t *prev = shaders[slot];
        if (prev == p){
	    /* shader is first list item */
            shaders[slot] = p->next;
        }
        else{
            while (prev != NULL){
                if (prev->next == p){
                    prev->next = p->next;
                    break;
                }
                prev = prev->next;
            }
        }
        /* free memory, allocated by shader_t */
        if (p->strings != NULL){
            for (i=0; i<p->nstring; i++){
                POV_FREE(p->strings[i]);
            }
            POV_FREE(p->strings);
        }
        for (i=0; i< HSIZE; i++){
            sym_t *p_sym=p->htab[i];
            sym_t *p_next;
            while (p_sym != NULL){
                p_next = p_sym->next;
                POV_FREE(p_sym->name);
                POV_FREE(p_sym);
                p_sym = p_next;
            }
        }
        POV_FREE(p->dimarr);
        POV_FREE(p->data_templ);
        POV_FREE(p->code);
        POV_FREE(p->name);
        POV_FREE(p->filename);
        POV_FREE(p);
    }
    POV_FREE(shdr);
}


#endif /* ShaderPatch */
